#include "as_parser.h"
#include "as_tokenizer.h"
#include "utils.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// BNF
// ===
//
// <prog> ::= <stmts>
// <stmts> ::= <stmt> newline | <stmt> newline <stmts>
// <stmt> ::= <label> <instr> | <instr> | <label>
// <instr> ::= <op> | <op> arg | <op> tag
// <label> ::= tag ":"
// <op> ::= "add" | "sub" | "mul" | "div" | "mod" | "eq" | ...

prog_t parse_prog(allocator_t alct, token_stream_t ts) {
    prog_t p = allocate(alct, sizeof(struct prog));
    p->stmts = parse_stmts(alct, ts);
    return p;
}

stmts_t parse_stmts(allocator_t alct, token_stream_t ts) {
    stmts_t ss = allocate(alct, sizeof(struct stmts));
    ss->stmts = allocate(alct, sizeof(stmt_t));
    ss->stmts[0] = NULL;
    int capacity = 0;
    int len = 0;
    while (peek_token(alct, ts)->type != TK_ENDOFFILE) {
        stmt_t s = parse_stmt(alct, ts);
        if (s == NULL) continue;
        if (len == capacity) {
            size_t new_capacity = capacity * 2 + 1;
            void* new_stmts = allocate(alct, sizeof(stmt_t*) * new_capacity);
            memcpy(new_stmts, ss->stmts, sizeof(stmt_t*) * capacity);
            ss->stmts = new_stmts;
            capacity = new_capacity;
        }
        // expect newline
        if (peek_token(alct, ts)->type == TK_NEWLINE) {
            next_token(alct, ts);
        } else {
            fprintf(stderr, "%d:%d expect newline.\n", peek_token(alct, ts)->line, peek_token(alct, ts)->col);
        }
        ss->stmts[len] = s;
        len++;
    }
    ss->stmts[len] = NULL;
    return ss;
}

label_t parse_label(allocator_t alct, token_stream_t ts) {
    token_t t = next_token(alct, ts);
    if (t->type != TK_TAG) {
        fprintf(stderr, "%d:%d expect label.\n", t->line, t->col);
        exit(-1);
    }
    label_t l = allocate(alct, sizeof(label_t));
    l->name = t->sval;
    t = next_token(alct, ts);
    if (t->type != TK_COLON) {
        fprintf(stderr, "%d:%d expect colon.\n", t->line, t->col);
        exit(-1);
    }
    return l;
}

stmt_t parse_stmt(allocator_t alct, token_stream_t ts) {
    token_t t = peek_token(alct, ts);
    stmt_t stmt = allocate(alct, sizeof(struct stmt));
    stmt->label = NULL;
    stmt->instr = NULL;
    if (t->type == TK_TAG) {
        stmt->label = parse_label(alct, ts);
        if (peek_token(alct, ts)->type == TK_NEWLINE) {
            return stmt;
        }
        t = peek_token(alct, ts);
    }
    if (t->type == TK_OP) {
        stmt->instr = parse_instr(alct, ts);
        if (peek_token(alct, ts)->type == TK_NEWLINE) {
            return stmt;
        }
    }
    if (t->type == TK_NEWLINE) {
        return NULL;
    }
    fprintf(stderr, "%d:%d expect lable + instruction, lable, or instruction.\n", t->line, t->col);
    exit(-1);
}

enum op parse_op(allocator_t alct, token_stream_t ts) {
    token_t t = next_token(alct, ts);
    enum op op;
    if (t->type == TK_OP) {
        op = str2op(t->sval);
        if (op == OP_END) {
            fprintf(stderr, "%d:%d invalid op.\n", t->line, t->col);
            exit(-1);
        }
    } else {
        fprintf(stderr, "%d:%d expect op.\n", t->line, t->col);
        exit(-1);
    }
    return op;
}

instr_t parse_instr(allocator_t alct, token_stream_t ts) {
    token_t t = peek_token(alct, ts);
    instr_t i = allocate(alct, sizeof(struct instr));
    i->tag_name = NULL;
    i->arg = NULL;
    i->op = OP_END;
    if (t->type == TK_OP) {
        i->op = parse_op(alct, ts);
        t = peek_token(alct, ts);
        if (t->type == TK_ARG) {
            arg_t a = allocate(alct, sizeof(struct arg));
            a->ival = t->ival;
            a->fval = t->fval;
            i->arg = a;
            next_token(alct, ts);
        } else if (t->type == TK_TAG) {
            i->tag_name = t->sval;
            next_token(alct, ts);
        }
    }
    return i;
}
